from helpers.http_calls import HttpCalls
import time
from getpass import getpass


class ZiaPortalTalker(object):
    """
    ZIA Portal API talker
    Documentation: https://help.zscaler.com/zia/zia-api/api-developer-reference-guide
    """

    def __init__(self, cloud_name):
        """
        Method to start the class
        :param cloud_name: type string. Example: zscalerbeta.net, zscalerone.net, zscalertwo.net
        zscalerthree.net, zscaler.net, zscloud.net
        """
        self.base_uri = f'https://admin.{cloud_name}/zsapi/v1'
        self.hp_http = HttpCalls(host=self.base_uri, verify=True)
        self.jsessionid = None
        self.zs_session_code = None
        self.headers = None
        self.version = '0.1'

    def _obfuscateApiKey(self, seed):
        """
        Internal method to Obfuscate the API key
        :param seed: API key
        :return: timestamp,obfuscated key
        """
        now = int(time.time() * 1000)
        n = str(now)[-6:]
        r = str(int(n) >> 1).zfill(6)
        key = ""
        for i in range(0, len(str(n)), 1):
            key += seed[int(str(n)[i])]
        for j in range(0, len(str(r)), 1):
            key += seed[int(str(r)[j]) + 2]
        return now, key

    def authenticate(self, apikey, username, password=None, ):
        """
        Method to authenticate.
        :param apikey: API key
        :param username: A string that contains the email ID of the API admin
        :param password: A string that contains the password for the API admin
        :return:  JSESSIONID. This cookie expires by default 30 minutes from last request
        """
        if not password:
            password = getpass(" Introduce password: ")
        timestamp, key = self._obfuscateApiKey('jj7tg80fEGao')

        payload = {
            "apiKey": key,
            "username": username,
            "password": password,
            "timestamp": timestamp
        }
        url = '/authenticatedSession'
        response = self.hp_http.post_call(url=url, payload=payload, headers={"Accept": "application/json"})
        if response.cookies.get('JSESSIONID'):
            self.jsessionid = response.cookies['JSESSIONID']
        else:
            raise ValueError("Invalid Credentials")
        if response.cookies.get('ZS_SESSION_CODE'):
            self.zs_session_code = response.cookies['ZS_SESSION_CODE']
            self.headers = {'Content-Type': 'application/json', 'ZS_CUSTOM_CODE': self.zs_session_code}
        else:
            raise ValueError("Invalid API key")

    def create_dlpEngine(self, payload=None, EngineExpression=None, Name=None, CustomDlpEngine=True,
                         PredefinedEngineName=None, Description=None):
        """
        Method to create a DLP engine
        :param Name: type string. Name of the DLP ENGINE
        :param EngineExpression: type string. Engine Expression
        :param CustomDlpEngine: : type boolean. True if custom DLP engine
        :param PredefinedEngineName: type boolean.
        :param Description: type string. Description
        :return: 
        """
        url = '/dlpEngines'
        if payload:
            payload = payload
        else:
            payload = {
                "EngineExpression": EngineExpression,
                "CustomDlpEngine": CustomDlpEngine
            }
            if PredefinedEngineName:
                payload.update(PredefinedEngineName=PredefinedEngineName)
            else:
                payload.update(Name=Name)

            if Description:
                payload.update(Description=Description)
        response = self.hp_http.post_call(url=url, payload=payload, headers=self.headers,
                                          cookies={'JSESSIONID': self.jsessionid,
                                                   'ZS_SESSION_CODE': self.zs_session_code,
                                                   })
        return response

    def update_dlpEngine(self, payload, id):
        """
        Method to update a DLP engine
        :param payload: type json. payload
        :return: 
        """
        url = f'/dlpEngines/{id}'
        response = self.hp_http.put_call(url=url, payload=payload, headers=self.headers,
                                         cookies={'JSESSIONID': self.jsessionid,
                                                  'ZS_SESSION_CODE': self.zs_session_code,
                                                  })
        return response

    def list_PacFiles(self):
        """
        Method to list PAC files
        :return: json
        """
        url = f'/pacFiles'
        response = self.hp_http.get_call(url=url, headers=self.headers,
                                         cookies={'JSESSIONID': self.jsessionid,
                                                  'ZS_SESSION_CODE': self.zs_session_code,
                                                  })
        return response.json()

    def add_PacFile(self, name, description, domain, PacContent, editable=True, pacUrlObfuscated=True):
        """
        Method to Add a PAC file
        :param name: type string. Name of the PAC
        :param description: type string. Description
        :param domain: type string. Domain
        :param PacContent: Type string: PAC content
        :param editable: Type boolean. Default True
        :param pacUrlObfuscated: Type boolean. Default True
        :return:
        """
        payload = {
            'name': name,
            'editable': editable,
            'pacContent': PacContent,
            'pacUrlObfuscated': pacUrlObfuscated,
            'domain': domain,
            'description': description,
            'pacVerificationStatus': 'VERIFY_NOERR'
        }
        url = f'/pacFiles'
        response = self.hp_http.post_call(url=url, headers=self.headers, payload=payload,
                                          cookies={'JSESSIONID': self.jsessionid,
                                                   'ZS_SESSION_CODE': self.zs_session_code,
                                                   })
        return response.json()



import pdb
import time

from helpers.http_calls import HttpCalls


class ZccTalker(object):
    """
    Client Connector API talker
    Documentation: under development
    Currently in beta status
    """

    def __init__(self, cloud):
        self.base_uri = f'https://api-mobile.{cloud}/papi'
        self.hp_http = HttpCalls(host=self.base_uri, verify=True)
        self.jsessionid = None
        self.version = 'beta 0.1'

    def authenticate(self, clientid, secretkey):
        """
        Method to authenticate.
        :param clientid: Client id
        :param secretkey: client secret, obtained from portal
        :return:  None
        """
        payload = {
            "apiKey": clientid,
            "secretKey": secretkey
        }
        url = '/auth/v1/login'
        response = self.hp_http.post_call(url=url, headers={'Accept': '*/*'}, payload=payload)
        self.header = {
            'auth-token': response.json()['jwtToken']
        }

    def _obtain_all(self, url, cookies=None, params=None, headers=None):
        """
        Internal method that queries all pages
        :param url:  URL
        :return:
        """
        page = 1
        result = []
        while True:
            response = self.hp_http.get_call(f'{url}&page={page}', cookies=cookies, params=params, headers=headers,
                                             error_handling=True)
            if response.json():
                result += response.json()
                page += 1
                time.sleep(0.5)
            else:
                break
        return result

    def list_devices(self, companyID, username=None, osType=None):
        """
        Method to authenticate.
        :param companyID: type int. ORG ID
        :param username: type string. Username in email format
        :param osType  type int.
        :return:  type list
        """
        if username:
            url = f'/public/v1/getDevices?username={username}'
        elif osType:
            url = f'/public/v1/getDevices?osType={osType}'
        else:
            url = '/public/v1/getDevices?pagesize100'
        response = self._obtain_all(url=url, params=companyID, headers=self.header)
        return response

    def list_OTP(self, companyID, udid):
        """
        Method to fetch the One Time Password for a specific device. These passwords are unique and tied to a device UDID
        :param companyID: type int. ORG ID
        :param udid: type int. User device ID
        :return: type list
        """
        url = f'/public/v1/getOtp?udid={udid}'
        response = self.hp_http.get_call(url=url, params=companyID, headers=self.header)
        return response.json()

    def list_passwords(self, companyID, udid):
        """
        Method to fetch the One Time Password for a specific device. These passwords are unique and tied to a device UDID
        :param companyID: type int. ORG ID
        :param udid: type int. User device ID
        :return: type list
        """
        url = f'/public/v1/getOtp?udid={udid}'
        response = self.hp_http.get_call(url=url, params=companyID, headers=self.header)
        return response.json()

    def remove_devices(self, companyID, udids):
        """
        Method to  mark the device for removal (Device Removal Pending).
        :param companyID: type int. ORG ID
        :param udids: type list. List of user devices ids
        :return: type list
        """
        url = f'/public/v1/removeDevices'
        payload = {"companyId": companyID,
                   "udids": udids
                   }
        response = self.hp_http.post_call(url=url, headers=self.header, payload=payload)
        return response.json()

    def force_remove_devices(self, companyID, udids):
        """
        Force Remove, has the same effect as Remove, though it additionally moves the device straight to Removed and also
        signals the cloud to invalidate the user’s session.
        :param companyID: type int. ORG ID
        :param udids: type list. List of user devices ids
        :return: type list
        """
        url = f'/public/v1/forceRemoveDevices'
        payload = {"companyId": companyID,
                   "udids": udids
                   }
        response = self.hp_http.post_call(url=url, headers=self.header, payload=payload)
        return response.json()



from helpers.http_calls import HttpCalls


class ZpaTalkerPublic(object):
    """
    ZPA API talker
    Documentation: https://help.zscaler.com/zpa/zpa-api/api-developer-reference-guide
    """

    def __init__(self, customerID, cloud='https://config.private.zscaler.com'):
        """
        :param cloud: type string. Example https://config.zpabeta.net
        :param customerID: type int. The unique identifier of the ZPA tenant
        """
        self.base_uri = f'{cloud}'
        # self.base_uri = f'https://config.zpabeta.net'
        self.hp_http = HttpCalls(host=self.base_uri, verify=True)
        self.jsessionid = None
        self.version = '1.3'
        self.header = None
        self.customerId = customerID

    def _obtain_all_results(self, url):
        """
        API response can have multiple pages. This method return the whole response in a list
        :param url: type string. url
        :return: type list
        """
        result = []
        if '?pagesize' not in url:
            url = url + '?pagesize=500'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        if 'list' not in response.json().keys():
            return []
        if int(response.json()['totalPages']) > 1:
            i = 0
            while i <= int(response.json()['totalPages']):
                result = result + \
                         self.hp_http.get_call(f'{url}&page={i}', headers=self.header, error_handling=True).json()[
                             'list']
                i += 1
        else:
            result = response.json()['list']
        return result

    def authenticate(self, client_id, client_secret):
        """
        Method to obtain the Bearer Token
        return token
        """
        url = f'/signin'
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        payload = {
            'client_id': client_id,
            'client_secret': client_secret
        }
        response = self.hp_http.post_call(url, headers=headers, error_handling=True, payload=payload, urlencoded=True)
        self.header = {
            'Authorization': f"{response.json()['token_type']} {response.json()['access_token']}"
        }
        return response.json()

    # app-server-controller

    def list_servers(self, query=False, serverId=None):
        """
        Method to obtain all the configured Servers.
        :param serverId: type int. Unique server id number
        :param query: type string. Example ?page=1&pagesize=20&search=consequat
        :return:json
        """
        if serverId:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/server/{serverId}'
        else:
            if not query:
                query = '?pagesize=500'
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/server{query}'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        return response.json()

    # application-controller
    def list_application_segments(self, applicationId=None):
        """
        Method to obtain application segments
        :param applicationId type int. Application unique identified id
        """
        if applicationId:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/application/{applicationId}'
            response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
            return response.json()
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/application'
        response = self._obtain_all_results(url)
        return response

    def add_application_segment(self, name, healthReporting, domainNames, segmentGroupId, serverGroups,
                                tcpPortRanges=[], udpPortRanges=[], description='', enabled=True, ipAnchored=False,
                                doubleEncrypt=False, bypassType='NEVER', isCnameEnabled=True, cnameConfig='NOFLATTEN'):
        """
        Adds a new Application Segment for a ZPA tenant.
        :param name: type string. App Name
        :param description: type string. Description
        :param enabled: type boolean (True|False)
        :param healthReporting: type string. possible values: NONE, ON_ACCESS, CONTINUOUS
        :param ipAnchored: type boolean (True|False)
        :param doubleEncrypt: type boolean (True|False)
        :param bypassType: type string. possible values ALWAYS, NEVER, ON_NET
        :param isCnameEnabled: type boolean (True|False)
        :param tcpPortRanges: type list.  ["from", "to"]
        :param udpPortRanges: type list.  ["from", "to"]
        :param domainNames: type list. List of domains or IP addresses
        :param segmentGroupId: type string. Application Segment Group id
        :param serverGroups=type list. list of dictionaries, where key is id and value is serverGroupId [{
                "id": "<serverGroupId>"}]
        :return: type dict. HTTP response
        """

        url = f"/mgmtconfig/v1/admin/customers/{self.customerId}/application"
        payload = {
            "name": name,
            "description": description,
            "enabled": enabled,
            "healthReporting": healthReporting,
            "ipAnchored": ipAnchored,
            "doubleEncrypt": doubleEncrypt,
            "bypassType": bypassType,
            "isCnameEnabled": isCnameEnabled,
            "tcpPortRanges": tcpPortRanges,
            "udpPortRanges": udpPortRanges,
            "domainNames": domainNames,
            "segmentGroupId": segmentGroupId,
            "serverGroups": serverGroups,
            "cnameConfig": cnameConfig
        }
        response = self.hp_http.post_call(url=url, payload=payload, headers=self.header, error_handling=True)
        return response.json()

    def update_application_segment(self,applicationId,payload):
        """
        Updates the Application Segment details for the specified ID
        :param applicationId: type int. Application ID
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customerId}/application/{applicationId}"
        response = self.hp_http.put_call(url=url, payload=payload, headers=self.header, error_handling=True)
        return response
    def delete_application_segment(self,applicationId,):
        """
        Updates the Application Segment details for the specified ID
        :param applicationId: type int. Application ID
        """
        url = f"/mgmtconfig/v1/admin/customers/{self.customerId}/application/{applicationId}"
        response = self.hp_http.delete_call(url=url, error_handling=True)
        return response
    # segment-group-controller

    def list_segment_group(self, segmentGroupId=None, query=False):
        """
        Get all the configured Segment Groups. If segmentGroupId obtains the segment sroup details
        :param segmentGroupId: The unique identifier of the Segment Group.
        :param query: url query: Example ?page=1&pagesize=20&search=consequat
        return json
        """
        if segmentGroupId:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/segmentGroup/{segmentGroupId}'
        else:
            if not query:
                query = '?pagesize=500'
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/segmentGroup{query}'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        return response.json()

    def add_segment_group(self, name, description, enabled=True):
        """
        Add a new segment group
        :param name: type string. Name of segment Group
        :param description: type string. Description
        :param enabled: type boolean: True or False
        :return: Json
        """
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/segmentGroup'
        payload = {
            "name": name,
            "description": description,
            "enabled": enabled,
        }
        response = self.hp_http.post_call(url, headers=self.header, error_handling=True, payload=payload)
        return response.json()

    # connector-controller
    def list_connector(self, connectorId=None, ):
        """
        Get all the configured Segment Groups. If segmentGroupId obtains the segment group details
        :param connectorId: The unique identifier of the App Connector.
        return json
        """
        if connectorId:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/connector/{connectorId}'
            return self.hp_http.get_call(url, headers=self.header, error_handling=True).json()
        else:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/connector'
        response = self._obtain_all_results(url)
        return response

    def delete_bulk_connector(self, ids):
        """
        Get all the configured Segment Groups. If segmentGroupId obtains the segment sroup details
        :param ids: type list. list of resouces ids for bulk deleting the App Connectors..
        return json
        """
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/connector/bulkDelete'
        payload = {"ids": ids
                   }
        response = self.hp_http.post_call(url=url, headers=self.header, error_handling=True, payload=payload)
        return response.json()

    # Connector-group-controller
    def list_connector_group(self, appConnectorGroupId=None):
        """
        Gets all configured App Connector Groups for a ZPA tenant.
        :param appConnectorGroupId: type int: The unique identifier of the Connector Group.
        return json
        """
        if appConnectorGroupId:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/appConnectorGroup/{appConnectorGroupId}'
            return self.hp_http.get_call(url, headers=self.header, error_handling=True).json()
        else:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/appConnectorGroup'
            response = self._obtain_all_results(url)
        return response

    # ba-certificate-controller-v-2

    def list_browser_access_certificates(self):
        """
        Get all Browser issued certificates
        return json
        """
        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/clientlessCertificate/issued'
        response = self._obtain_all_results(url)
        return response

    # enrollment-cert-controller

    def list_enrollment_certificates(self, ):
        """
        Get all the Enrollment certificates
        return list
        """
        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/enrollmentCert'
        response = self._obtain_all_results(url)
        return response

    def list_browser_access_certificates(self):
        """
        Get all the issued certificates
        return list
        """
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/visible/versionProfiles'
        response = self._obtain_all_results(url)
        return response

    # customer-version-profile-controller

    def list_customer_version_profile(self, query=False):
        """
        Get Version Profiles visible to a customer
        :param query: url query: Example ?page=1&pagesize=20&search=consequat
        return json
        """
        if not query:
            query = '?pagesize=500'
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/visible/versionProfiles{query}'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        return response.json()

    # cloud - connector - group - controller
    def list_cloud_connector_group(self, id=None, query=False):
        """
        Get all configured Cloud Connector Groups. If id, Get the Cloud Connector Group details
        :param query: url query: Example ?page=1&pagesize=20&search=consequat
        return json
        """
        if id:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/cloudConnectorGroup/{id}'
        else:
            if not query:
                query = '?pagesize=500'
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/cloudConnectorGroup{query}'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        return response.json()

    # idp-controller-v-2
    def list_idP(self, query=False):
        """
        Method to Get all the idP details for a ZPA tenant
        :param query: HTTP query
        :return: json
        """
        if not query:
            query = '?pagesize=500'

        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/idp{query}'
        response = self._obtain_all_results(url)
        return response

    # provisioningKey-controller
    def list_provisioningKey(self, associationType='CONNECTOR_GRP'):
        """
        Gets details of all the configured provisioning keys.
        :param associationType: type string. The supported values are CONNECTOR_GRP and SERVICE_EDGE_GRP.
        :return: list
        """
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/associationType/{associationType}/provisioningKey'
        response = self._obtain_all_results(url)
        return response

    # policy-set-controller

    # scim-attribute-header-controller

    def list_scim_attributes(self, idpId, query=False):
        """

        :param idpId: The unique identifies of the Idp
        :param query: ?page=1&pagesize=20&search=consequat
        """
        if not query:
            query = '?pagesize=500'
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/idp/{idpId}/scimattribute{query}'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        return response.json()

    # scim-group-controller
    def list_scim_groups(self, idpId, query=False):
        """
        Method to list all SCIM groups
        :param idpId: The unique identifies of the Idp
        :param query: ?page=1&pagesize=20&search=consequat
        """
        if not query:
            query = '?pagesize=500'
        url = f'/userconfig/v1/customers/{self.customerId}/scimgroup/idpId/{idpId}{query}'
        response = self._obtain_all_results(url)
        return response

    # saml-attr-controller-v-2
    def list_saml_attributes(self):
        """
        Method to get all SAML attributes
        """
        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/samlAttribute'
        response = self._obtain_all_results(url)
        return response

    # global-policy-controller

    def list_policies(self, policyType='ACCESS_POLICY'):
        """list policie(s)  by policy type,
         :param policyType: Type string. Supportef values Possible values=ACCESS_POLICY,GLOBAL_POLICY, TIMEOUT_POLICY,REAUTH_POLICY,
         SIEM_POLICY, CLIENT_FORWARDING_POLICY,BYPASS_POLICY
         """
        url = f"/mgmtconfig/v1/admin/customers/{self.customerId}/policySet/rules/policyType/{policyType}"
        response = self._obtain_all_results(url)
        return response

    def list_global_policy_id(self, query=False):
        """
        Method to get the global policy
        :param query: ?page=1&pagesize=20&search=consequat
        """
        if not query:
            query = '?pagesize=500'
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/policySet/global{query}'
        response = self.hp_http.get_call(url, headers=self.header, error_handling=True)
        return response.json()

    def add_policySet(self, app_operands, RuleName, Action, policySetId, operands, operator, MsgString=None):
        """
        Method to create a new access Policy
        :param app_operands: list of app_operands: Examples
        [{
                    "objectType": "APP",
                    "lhs": "id",
                    "rhs": applicationId,
        }]
        :param RuleName: Policy set Rule Name
        :param Action: ALLOW / DENY
        :param policySetId:  Global Policy ID. can be obtained from list_global_policy_id
        :param operands:  List of operands. Example
        [{
            "objectType": "SAML",
            "lhs": "<samlAttrId>",
            "rhs": "<samlAttrValue>",
        },{
            "objectType": "SCIM",
            "lhs": "<scimAttrId>",
            "rhs": "<scimAttrValue>”
        }]


        :return:
        """
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/policySet/{policySetId}/rule'
        payload = {
            "conditions": [{
                "operands": app_operands
            }, {

                "operands": operands,
                "operator": operator,
            }, ],
            # Seems here needs to be AND
            "operator": 'AND',
            "name": RuleName,
            "description": "Description",
            "action": Action,
            "customMsg": MsgString
        }
        print(payload)
        response = self.hp_http.post_call(url=url, headers=self.header, error_handling=True, payload=payload)
        return response.json()

    # Server Group Controller

    def list_server_groups(self, groupId=None):
        """
        Method to get all configured Server Groups. If groupI, get the Server Group details
        :param groupId: type integer. The unique identifier of the Server Group.
        return json
        """
        if groupId:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/serverGroup/{groupId}'
            response = self.hp_http.get_call(url, headers=self.header, error_handling=True).json()
        else:
            url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/serverGroup'
            response = self._obtain_all_results(url)
        return response

    def add_server_groups(self, name, description, connector_group_id):
        """
        :param name: Server Group Name
        :param description: Description
        :param connector_group_id: list of dictionaries with key as "id" and value connector_group_id.
            [{"id": connector_group_id}]
        """
        url = f'/mgmtconfig/v1/admin/customers/{self.customerId}/serverGroup'

        payload = {
            "enabled": True,
            "dynamicDiscovery": True,
            "name": name,
            "description": description,
            "servers": [
            ],
            "appConnectorGroups": connector_group_id
        }
        response = self.hp_http.post_call(url=url, headers=self.header, error_handling=True, payload=payload)
        return response.json()

    def list_posture_profiles(self, query=False):
        """
        Method to Get all the idP details for a ZPA tenant
        :param query: HTTP query
        :return: json
        """
        if not query:
            query = '?pagesize=500'

        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/posture{query}'
        response = self._obtain_all_results(url)
        return response

    def list_privileged_consoles(self, query=False):
        """
        Method to Get all the privleged_remote_consoles for a ZPA tenant
        :param query: HTTP query
        :return: json
        """
        if not query:
            query = '?pagesize=500'

        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/privilegedConsoles{query}'
        response = self._obtain_all_results(url)
        return response

    def list_sra_consoles(self):
        """
        Method to obtain list of sra consoles from all application segments
        """
        sralist = []
        appsegments = self.list_application_segments()
        for apps in appsegments:
            srap = apps.get("sraApps")
            if srap is not None:
                sralist.extend(srap)
        return sralist

    # Certificate Controller v2
    def list_issued_certificates(self, query=None):
        """
        Method to get all issued certificates
        """
        if not query:
            query = '?pagesize=500'

        url = f'/mgmtconfig/v2/admin/customers/{self.customerId}/certificate/issued'
        response = self._obtain_all_results(url)
        return response



import pdb

from helpers.http_calls import HttpCalls
import time
from getpass import getpass
from models.models import valid_category_ids
from models.models import super_categories
from models.models import valid_countries


class ZiaTalker(object):
    """
    ZIA API talker
    Documentation: https://help.zscaler.com/zia/zia-api/api-developer-reference-guide
    """

    def __init__(self, cloud_name):
        """
        Method to start the class
        :param cloud_name: type string. Example: zsapi.zscalerbeta.net, zsapi.zscalerone.net, zsapi.zscalertwo.net
        zsapi.zscalerthree.net, zsapi.zscaler.net, zsapi.zscloud.net
        """
        self.base_uri = f'https://{cloud_name}/api/v1'
        self.hp_http = HttpCalls(host=self.base_uri, verify=True)
        self.jsessionid = None
        self.version = '1.2'

    def _obfuscateApiKey(self, seed):
        """
        Internal method to Obfuscate the API key
        :param seed: API key
        :return: timestamp,obfuscated key
        """
        now = int(time.time() * 1000)
        n = str(now)[-6:]
        r = str(int(n) >> 1).zfill(6)
        key = ""
        for i in range(0, len(str(n)), 1):
            key += seed[int(str(n)[i])]
        for j in range(0, len(str(r)), 1):
            key += seed[int(str(r)[j]) + 2]
        return now, key

    def authenticate(self, apikey, username, password=None, ):
        """
        Method to authenticate.
        :param apikey: API key
        :param username: A string that contains the email ID of the API admin
        :param password: A string that contains the password for the API admin
        :return:  JSESSIONID. This cookie expires by default 30 minutes from last request
        """
        if not password:
            password = getpass(" Introduce password: ")
        timestamp, key = self._obfuscateApiKey(apikey)

        payload = {
            "apiKey": key,
            "username": username,
            "password": password,
            "timestamp": timestamp
        }
        url = '/authenticatedSession'
        response = self.hp_http.post_call(url=url, payload=payload)
        self.jsessionid = response.cookies['JSESSIONID']

    def authenticated_session(self):
        """
        Checks if there is an authenticated session
        :return: json
        """
        url = '/authenticatedSession'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True, )
        print(response)
        return response.json()

    def end_session(self):
        """
        Menthod to enf an authenticated session
        :return: None
        """
        url = '/authenticatedSession'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True,
                                            payload={})
        return response.json()

    def _obtain_all(self, url):
        """
        Internal method that queries all pages
        :param url:  URL
        :return:
        """
        page = 1
        result = []
        while True:
            response = self.hp_http.get_call(f'{url}&page={page}', cookies={'JSESSIONID': self.jsessionid},
                                             error_handling=True)
            if response.json():
                result += response.json()
                page += 1
                time.sleep(1)
            else:
                break
        return result

    def get_status(self):
        """
        Method to obtain the activation status for a configuration change
        :return: json object with the status
        """
        url = '/status'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True)
        return response.json()

    def activate_status(self):
        """
        Method to activate configuration changes
        :return: json object with the status
        """
        url = '/status/activate'
        response = self.hp_http.post_call(url, payload={}, cookies={'JSESSIONID': self.jsessionid}, error_handling=True)
        return response.json()

    # Admin Audit Logs

    def list_auditlogEntryReport(self):
        """
        Gets the status of a request for an audit log report. After sending a POST request to /auditlogEntryReport to
        generate a report, you can continue to call GET /auditlogEntryReport to check whether the report has finished
        generating. Once the status is COMPLETE, you can send another GET request to /auditlogEntryReport/download to
        download the report as a CSV file.
        :return: json
        """

        url = "/auditlogEntryReport"

        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def download_auditlogEntryReport(self):
        """
        Gets the status of a request for an audit log report. After sending a POST request to /auditlogEntryReport to
        generate a report, you can continue to call GET /auditlogEntryReport to check whether the report has finished
        generating. Once the status is COMPLETE, you can send another GET request to /auditlogEntryReport/download to
        download the report as a CSV file.
        :return: json
        """

        url = "/auditlogEntryReport/download"
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response

    def add_auditlogEntryReport(self, startTime, endTime, actionTypes=None, category=None,
                                subcategories=None, actionInterface=None, ):
        """
         Creates an audit log report for the specified time period and saves it as a CSV file. The report includes audit
         information for every call made to the cloud service API during the specified time period.
         Creating a new audit log report will overwrite a previously-generated report.
        :param startTime: The timestamp, in epoch, of the admin's last login
        :param endTime: The timestamp, in epoch, of the admin's last logout.
        :param actionTypes: type list. The action performed by the admin in the ZIA Admin Portal or API
        :param actionResult: The outcome (i.e., Failure or Success) of an actionType.
        :param category: tyoe string. The location in the Zscaler Admin Portal (i.e., Admin UI) where the actionType was performed
        :param subcategories: type list. The area within a category where the actionType was performed.
        :param actionInterface: type string. The interface (i.e., Admin UI or API) where the actionType was performed.
        :param clientIP: type string. The source IP address for the admin
        :param adminName: type string.  The admin's login ID
        :return: 204 Successfull Operation
        """
        url = "/auditlogEntryReport"
        payload = {"startTime": startTime,
                   "endTime": endTime,
                   }
        if category:
            payload.update(category=category)
        if subcategories:
            payload.update(subcategories=subcategories)
        if actionInterface:
            payload.update(actionInterface=actionInterface)
        if actionTypes:
            payload.update(actionTypes=actionTypes)

        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response

    # Admin & Role Management
    def list_adminUsers(self, userId=None, query=None):
        """
        Gets a list of admin users. By default, auditor user information is not included.
        :param userId: user ID
        :param query: HTTP query
        :return:json()
        """
        if userId:
            url = f'/adminUsers/{userId}'
            return self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True).json()
        else:
            if query:
                url = f"/adminUsers?{query}?pageSize=1000"
            else:
                url = "/adminUsers?pageSize=1000"
        return self._obtain_all(url)

    def list_adminRoles(self, query=None):
        """
        Gets a name and ID dictionary of al admin roles
        :param query: HTTP query
        :return: json
        """
        if query:
            url = f"/adminRoles/lite?{query}"
        else:
            url = "/adminRoles/lite"
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    # URL Categories
    def list_url_categories(self, custom=False):
        """
        Gets information about all or custom URL categories
        :param custom: Boolean, if True it will return custom categories only
        :return: json
        """

        if custom:
            url = '/urlCategories?customOnly=true'
        else:
            url = '/urlCategories'
        #return self._obtain_all(url)
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True)
        return response.json()

    def list_url_categories_lite(self):
        """
        Gets a lightweight key-value list of all or custom URL categories.
        """
        url = '/urlCategories/lite'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True)
        return response.json()

    def add_url_categories(self, name, superCategory, type='URL_CATEGORY', urls=None, dbCategorizedUrls=None,
                           keywordsRetainingParentCategory=[], keywords=[], customCategory=False, ipRanges=[],
                           ipRangesRetainingParentCategory=[]):
        """
         Adds a new custom URL category.
        :param name: type string. Name of the custom category. Possible values URL_CATEGORY, TLD_CATEGORY, ALL
        :param superCategory: type string. super category
        :param urls: type list. List of urls
        :param dbCategorizedUrls: type list. URL retaining parent category
        :param keywordsRetainingParentCategory: type list. Retained custom keywords from the parent URL category that is associated to a URL category.
        :param keywords: type list. Custom keywords associated to a URL category.
        :param customCategory: type boolean. Default False. Set to True for custom category
        :param ipRanges: type list. Custom IP address ranges associated to a URL category
        :param ipRangesRetainingParentCategory: The retaining parent custom IP address ranges associated to a URL category.

        :return:  json
        """
        if keywordsRetainingParentCategory is None:
            keywordsRetainingParentCategory = []

        if superCategory not in super_categories:
            print(f'Error -> Invalid Super Category')
            print(f'{super_categories}')
            raise ValueError("Invalid super category")

        url = '/urlCategories'
        payload = {
            "configuredName": name,
            "customCategory": customCategory,
            "superCategory": superCategory,
            "keywordsRetainingParentCategory": keywordsRetainingParentCategory,
            "keywords": keywords,
            "urls": urls,
            "dbCategorizedUrls": dbCategorizedUrls,
            "ipRanges": ipRanges,
            "ipRangesRetainingParentCategory": ipRangesRetainingParentCategory,
            "type": type
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def add_url_categories1(self, payload):
        """
         Adds a new custom URL category.
        :param payload
        :return:  json
        """
        url = '/urlCategories'
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def update_url_categories(self, categoryId, action=None, configuredName=None, urls=None, dbCategorizedUrls=None,
                              keywords=None, keywordsRetainingParentCategory=None, ):
        """
         Updates the URL category for the specified ID. If keywords are included within the request, then they
         will replace existing ones for the specified URL category . If the keywords attribute is not included the
         request, the existing keywords are retained.
         You can perform a full update for the specified URL category. However, if attributes are omitted within the
          update request, the values for those attributes are cleared.

         You can also perform an incremental update, to add or remove URLs, for the specified URL category using the
         action parameter
         :param categoryId: type string. URL id
        :param configuredName: type string. Name of the custom category
        :param urls: list of urls
        "param dbCategorizedUrls: type list. URL retaining parent category
        :param keywordsRetainingParentCategory: list of key works
        :param action: Optional parameter. ADD_TO_LIST or REMOVE_FROM_LIST
        :return:  json
        """
        '''if categoryId not in valid_category_ids:
            print(f'Error -> Invalid category id')
            raise ValueError("Invalid category id")'''

        if action == 'ADD_TO_LIST':
            url = f'/urlCategories/{categoryId}?action=ADD_TO_LIST'
        elif action == 'REMOVE_FROM_LIST':
            url = f'/urlCategories/{categoryId}?action=REMOVE_FROM_LIST'
        elif not action:
            url = f'/urlCategories/{categoryId}'
        else:
            print(f'Error -> Invalid action')
            print(f'{action}')
            raise ValueError("Invalid action")

        payload = {
            "configuredName": configuredName,

        }
        if keywordsRetainingParentCategory:
            payload.update(keywordsRetainingParentCategory=keywordsRetainingParentCategory)
        if keywords:
            payload.update(keywords=keywords)
        if configuredName:
            payload.update(configuredName=configuredName)
        if urls:
            payload.update(urls=urls)
        if dbCategorizedUrls:
            payload.update(dbCategorizedUrls=dbCategorizedUrls)

        response = self.hp_http.put_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def delete_url_categories(self, categoryid):
        """
        Deletes the custom URL category for the specified ID.
        You cannot delete a custom category while it is being used by a URL policy or NSS feed. Also, predefined
        categories cannot be deleted.
        :param categoryid: Category ID
        :return: json response
        """
        url = f'/urlCategories/{categoryid}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response

    def delete_urlFilteringRules(self, ruleId):
        """
        Deletes the custom URL category for the specified ID.
        You cannot delete a custom category while it is being used by a URL policy or NSS feed. Also, predefined
        categories cannot be deleted.
        :param ruleId:  type int. Rule Id
        :return: json response
        """
        url = f'/urlFilteringRules/{ruleId}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response

    def list_url_categories_urlquota(self):
        """
        Gets information on the number of unique URLs that are currently provisioned for your organization as well as
        how many URLs you can add before reaching that number.
        :return: json
        """
        url = '/urlCategories/urlQuota'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        print(response.json())
        return response.json()

    def list_url_categories_id(self, category_id):
        """
        Gets the URL category information for the specified ID
        :param category_id:
        :return:
        """

        url = f"/urlCategories/{category_id}"

        if category_id not in valid_category_ids:
            print(f'Error -> Invalid Category ID')
            print(f'{valid_category_ids}')
            raise ValueError("Invalid Category ID")
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def url_lookup(self, url_list):
        """
        Method to look up the categorization of the given list of URLs, ["abc.com","zyz.com"]
        :param url_list: list of urls
        :return:  json
        """
        result = []
        url = '/urlLookup'
        # Verify urls format
        list(set(url_list))
        # Rate limit 1/sec  and 400 hr and 100 URLs per call
        list_of_lists = [url_list[i:i + 100] for i in range(0, len(url_list), 100)]
        for item in list_of_lists:
            response = self.hp_http.post_call(url, payload=item, cookies={'JSESSIONID': self.jsessionid},
                                              headers={'Connection': 'close'},
                                              error_handling=True)
            result.append(response.json())
            time.sleep(1)
        final_result = []
        for i in result:
            for j in i:
                final_result.append(j)
        return final_result

    # URL filtering Policies
    def list_url_filtering_rules(self, ):
        """
        Gets a list of all of URL Filtering Policy rules
        :return:
        """
        url = '/urlFilteringRules'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_url_filtering_rules(self, name, order, protocols, state,
                                action, urlcategories=[], requestMethods=None, description=None, groups=None,
                                locations=None, departments=None, users=None, rank=7, locationGroups=None,
                                enforceTimeValidity=False,
                                validityEndTime=None, validityStartTime=None, validityTimeZoneId=None, cbiProfileId=0,
                                blockOverride=False):
        """
         Adds a URL Filtering Policy rule.
         If you are using the Admin Rank feature, refer to About Admin Rank to determine which value to provide for rank
         when adding a policy rule. If you are not using Admin Rank, the rank value must be 7.
        :param name: type string.  Name of the rule
        :param order: type integer. Rule order
        :param protocols: type string. Possible values SMRULEF_ZPA_BROKERS_RULE, ANY_RULE, TCP_RULE, UDP_RULE, DOHTTPS_RULE, TUNNELSSL_RULE,
        HTTP_PROXY, FOHTTP_RULE, FTP_RULE, HTTPS_RULE, HTTP_RULE, SSL_RULE, TUNNEL_RULE
        :param locations: type list. Each element is a  dictionary: Name-ID pairs of locations for which rule must be applied
        :param groups: type list. Name-ID pairs of groups for which rule must be applied
        :param departments:type list. Name-ID pairs of departments for which rule will be applied
        :param users: type list. Name-ID pairs of users for which rule must be applied
        :param urlcategories: type list. List of URL categories for which rule must be applied
        :param admin_rack:Admin rank of the admin who creates this rule
        :param timewindows: type list. Name-ID pairs of time interval during which rule must be enforced.
        :param requestmethods: type list. Request method for which the rule must be applied. If not set, rule will be applied to all
         methods
        :param endUserNotificationUrl: type string. URL of end user notification page to be displayed when the rule is matched. Not applicable if either
        'overrideUsers' or 'overrideGroups' is specified.
        :param overrideusers: Name-ID pairs of users for which this rule can be overridden. Applicable only if
         blockOverride is set to 'true', action is 'BLOCK' and overrideGroups is not set.If this overrideUsers is not
         set, 'BLOCK' action can be overridden for any user.
        :param overridegroups: Name-ID pairs of groups for which this rule can be overridden. Applicable only if
        blockOverride is set to 'true' and action is 'BLOCK'. If this overrideGroups is not set, 'BLOCK' action can be
        overridden for any group
        :param blockOverride: boolean: When set to true, a 'BLOCK' action triggered by the rule could be overridden.
        If true and both overrideGroup and overrideUsers are not set, the BLOCK triggered by this rule could be
        overridden for any users. If blockOverride is not set, 'BLOCK' action cannot be overridden.
        :param description: Additional information about the URL Filtering rule
        :param state: enabled/disabled
        :param action: Allow, Caution, Block
        :return:
        """
        url = '/urlFilteringRules'
        payload = {
            "blockOverride": blockOverride,
            "cbiProfileId": cbiProfileId,
            "description": description,
            "enforceTimeValidity": enforceTimeValidity,
            "name": name,
            "order": order,
            "protocols": protocols,
            "urlCategories": urlcategories,
            "state": state,
            "rank": rank,
            "action": action
        }
        if locations:
            payload.update(locations=locations)
        if locationGroups:
            payload.update(locationGroups=locationGroups)
        if groups:
            payload.update(groups=groups)
        if departments:
            payload.update(departments=departments)
        if users:
            payload.update(users=users)
        if requestMethods:
            payload.update(requestMethods=requestMethods)
        if enforceTimeValidity:
            payload.update(validityStartTime=validityStartTime)
            payload.update(validityEndTime=validityEndTime)
            payload.update(validityTimeZoneId=validityTimeZoneId)

        print(payload)

        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    # User Management

    def list_departments(self, department_id=""):
        """
        Gets a list of departments. The search parameters find matching values within the "name" or "comments"
        attributes.
        if ID, gets the department for the specified ID
        :param department_id: department ID
        :return:json()
        """

        if department_id:
            url = "/departments"
        else:
            url = f'/departments/{department_id}'

        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_groups(self, group_id=None):
        """
        Gets a list of groups
        if ID, gets the group for the specified ID
        :param group_id: group ID
        :return:json()
        """
        if not group_id:
            url = "/groups?pageSize=10000"
            return self._obtain_all(url)
        else:
            url = f'/groups/{group_id}'
            response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                             error_handling=True)

        return response.json()

    def list_users(self, user_id=None, query=None):
        """
        Gets a list of all users and allows user filtering by name, department, or group.
        The name search parameter performs a partial match. The dept and group parameters perform a 'starts with' match.
        if ID, gets user information for the specified ID
        :param user_id: user ID
        :return:json()
        """
        if user_id:
            url = f'/users/{user_id}'
            return self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True).json()
        else:
            if query:
                url = f"/users?{query}&pageSize=1000"
                return self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                             error_handling=True).json()
            else:
                url = "/users?pageSize=1000"
        return self._obtain_all(url)

    def add_users(self, name, email, groups, department, comments, password, adminuser=False, ):
        """
        Adds a new user. A user can belong to multiple groups, but can only belong to one department.

        :param name: string, user name
        :param email: string user email address
        :param groups: list. each member is a dictionary, key id, value name [{"id":1234, "name":"guest-wifi"}]
        :param department: dictionary, key is the id and value is the name {"id":1234, "name":"guests"}
        :param comments: string, comments
        :param password: string password,
        :param adminuser: True if user is admin user. default False
        :return: Json
        """
        url = '/users'
        payload = {"name": name,
                   "email": email,
                   "groups": groups,
                   "department": department,
                   "comments": comments,
                   "password": password,
                   "adminUser": adminuser}
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def delete_bulk_users(self, user_ids):
        """
        Bulk delete users up to a maximum of 500 users per request. The response returns the user IDs that were
        successfully deleted.
        :param user_ids:  List of user IDS to be deleted
        """
        url = '/users/bulkDelete'
        if len(user_ids) < 500:
            payload = {
                "ids": user_ids
            }
            response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                              error_handling=True)
            return response.json()
        else:
            raise ValueError("Maximum 500 users per request")

    # Location Management

    def list_locations(self, locationId=None):
        """
        Gets locations only, not sub-locations. When a location matches the given search parameter criteria only its
        parent location is included in the result set, not its sub-locations.
        :param locationId: Location id
        """
        if locationId:
            url = f'/locations/{locationId}'
        else:
            url = f'/locations'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_sublocations(self, locationId):
        """
        Gets the sub-location information for the location with the specified ID
        :param locationId: Location id
        """
        if locationId:
            url = f'/locations/{locationId}/sublocations'
        else:
            url = f'/locations'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_locationsgroups(self, ):
        """
        Gets information on location groups
        :param locationgroupId: Location group id
        """
        url = f'/locations/groups'
        print(url)
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def delete_bulk_locations(self, locationIds):
        """
        Bulk delete locations up to a maximum of 100 users per request. The response returns the location IDs that were successfully deleted..
        :param locationIds: list of location IDs
        """
        url = '/locations/bulkDelete'
        if len(locationIds) < 100:
            payload = {
                "ids": locationIds
            }
            response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                              error_handling=True)
            return response.json()
        else:
            raise ValueError("Maximum 100 locations per request")

    def delete_locations(self, locationId):
        """
        Deletes the location or sub-location for the specified ID
        :param locationId: location ID
        """
        url = f'/locations/{locationId}'

        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response

    #   Traffic Forwarding

    def list_greTunnels(self, greTunnelId=None):
        """
        Gets the GRE tunnel information for the specified ID
        :param greTunnelId: Optional. The unique identifier for the GRE tunnel
        """
        if greTunnelId:
            url = f'/greTunnels/{greTunnelId}'
        else:
            url = f'/greTunnels'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_greTunnels(self, sourceIp, primaryDestVip, secondaryDestVip, internalIpRange,
                       withinCountry, comment, ipUnnumbered):
        """
        Adds a GRE tunnel configuration.
        :param sourceIp: type string. The source IP address of the GRE tunnel. This is typically a static IP address in
         the organization or SD-WAN. This IP address must be provisioned within the Zscaler service using the /staticIP
         endpoint.
        :param primaryDestVip: type dictionary. {id:value} where value is integer: Unique identifier of the GRE primary
         VIP
        :param secondaryDestVip: type dictionary. {id:value} where value is integer: Unique identifier of the GRE
        secondary VIP
        :param internalIpRange: type string. The start of the internal IP address in /29 CIDR range
        :param withinCountry: type boolean. Restrict the data center virtual IP addresses (VIPs) only to those within
        the same country as the source IP address
        :param comment: type string. Additional information about this GRE tunnel
        :param ipUnnumbered:This is required to support the automated SD-WAN provisioning of GRE tunnels, when set to
        true gre_tun_ip and gre_tun_id are set to null
        :return:
        """
        url = f'/greTunnels'
        payload = {
            "sourceIp": sourceIp,
            "primaryDestVip": primaryDestVip,
            "secondaryDestVip": secondaryDestVip,
            "internalIpRange": internalIpRange,
            "withinCountry": withinCountry,
            "comment": comment,
            "ipUnnumbered": ipUnnumbered
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def list_gre_validateAndGetAvailableInternalIpRanges(self):
        """
        Gets the next available GRE tunnel internal IP address ranges
        :return: list of available IP addresses
        """
        url = f'/greTunnels/availableInternalIpRanges'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_gre_recommended_vips(self, query):
        """
        Gets a list of recommended GRE tunnel virtual IP addresses (VIPs),
        based on source IP address or latitude/longitude coordinates.
        :param query: type string. URL query. Example:
        :return: list of available IP addresses
        """
        url = f'/vips/recommendedList?{query}'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_gre_validate_ip(self, ip):
        """
        Gets the static IP address and location mapping information for the specified GRE tunnel
        IP address
        :param ip: type string. IP address of the GRE tunnel.
        :return:
        """
        url = f'/greTunnels/validateIP/{ip}'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_vpnCredentials(self, vpnId=None):
        """
        Gets VPN credentials that can be associated to locations.
        :param vpnId: Optional. If specified, get VPN credentials for the specified ID.
        """
        if vpnId:
            url = f'/vpnCredentials/{vpnId}'
        else:
            url = f'/vpnCredentials'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_vpnCredentials(self, fqdn, preSharedKey, type='UFQDN', comments=None, ):
        """
        Adds VPN credentials that can be associated to locations.
        :param fqdn: type string. Example abc@domain.com
        :param preSharedKey: type string. Pre-shared key. This is a required field for UFQDN and IP auth type
        :param type: type string VPN authentication type. valid options CN, IP, UFQDN,XAUTH
        :param comments: type string. Additional information about this VPN credential.
        :return:
        """
        url = f'/vpnCredentials'
        payload = {
            "type": type,
            "fqdn": fqdn,
            "preSharedKey": preSharedKey
        }
        if comments:
            payload.update(comments=comments)
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def delete_vpnCredentials(self, vpnId):
        """
        Deletes the VPN credentials for the specified ID.
        :param vpnId: type integer. The unique identifier for the VPN credential.
        :return:
        """
        url = f'/vpnCredentials/{vpnId}'

        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response

    def list_staticIP(self, IPId=None):
        """
        Gets all provisioned static IP addresses.
        :param IPId: Optional. If specified, get IP address for the specified id
        """
        if IPId:
            url = f'/staticIP/{IPId}'
        else:
            url = f'/staticIP'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_staticIP(self, ipAddress, geoOverrride=False, routableIP=True, latitude=0, longitude=0, comment=''):
        """
        Adds a static IP address
        :param ipAddress: String. The satic IP address
        :param geoOverrride: Boolean. If not set, geographic coordinates and city are automatically determined from
        the IP address. Otherwise, the latitude and longitude coordinates must be provided.
        :param routableIP: Boolean. Indicates whether a non-RFC 1918 IP address is publicly routable. This attribute is
        ignored if there is no ZIA Private Service Edge associated to the organization.
        :param latitude: Required only if the geoOverride attribute is set. Latitude with 7 digit precision after
        decimal point, ranges between -90 and 90 degrees.
        :param longitude: Required only if the geoOverride attribute is set. Longitude with 7 digit precision after
        decimal point, ranges between -180 and 180 degrees.
        :param comment: String Additional information about this static IP address
        """
        url = '/staticIP'

        payload = {
            "ipAddress": ipAddress,
            "latitude": latitude,
            "longitude": longitude,
            "routableIP": routableIP,
            "comment": comment
        }
        if geoOverrride:
            payload.update(geoOverrride=geoOverrride)
            payload.update(latitude=latitude)
            payload.update(longitude=longitude)

        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def delete_staticIP(self, Id):
        """
        Deletes the static IP address for the specified ID.
        :param Id: type integer. The unique identifier for the provisioned static IP address.
        :return: json
        """
        url = f'/staticIP/{Id}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)

        return response

    # User Authentication Settings
    def list_exemptedUrls(self):
        """
        Gets a list of URLs that were exempted from cookie authentication
        """
        url = '/authSettings/exemptedUrls'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_exemptedUrls(self, urls):
        """
        Adds URLs to the cookie authentication exempt list to the list
        :param urls: List of urls. Example ['url1','url2']
        """
        url = '/authSettings/exemptedUrls?action=ADD_TO_LIST'
        payload = {"urls": urls}
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def delete_exemptedUrls(self, urls):
        """
        Removed URLs to the cookie authentication exempt list to the list
        :param urls: List of urls. Example ['url1','url2']
        """
        url = '/authSettings/exemptedUrls?action=REMOVE_FROM_LIST'
        payload = {"urls": urls}
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    # Security Policy Settings

    def list_security_whitelisted_urls(self):
        """
        Gets a list of white-listed URLs
        """
        url = '/security'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def update_security_whitelisted_urls(self, urls):
        """
        Updates the list of white-listed URLs. This will overwrite a previously-generated white list.
        If you need to completely erase the white list, submit an empty list.
        :param urls: list of urls ['www.zscaler.com', 'www.example.com']
        """
        url = '/security'
        payload = {
            "whitelistUrls": urls
        }
        response = self.hp_http.put_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_security_blacklisted_urls(self):
        """
        Gets a list of white-listed URLs
        """
        url = '/security/advanced'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def update_security_blacklisted_urls(self, urls):
        """
       Updates the list of black-listed URLs. This will overwrite a previously-generated black list.
       If you need to completely erase the black list, submit an empty list.
        """
        url = '/security/advanced'
        payload = {
            "blacklistUrls": urls
        }
        response = self.hp_http.put_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_security_blacklistUrls(self, urls):
        """
        :param urls: list of urls
        Adds a URL from the black list. To add a URL to the black list.
        """
        url = '/security/advanced/blacklistUrls?action=ADD_TO_LIST'
        payload = {
            "blacklistUrls": urls
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response

    def remove_security_blacklistUrls(self, urls):
        """
        Removes a URL from the black list.
        :param urls: List of urls
        """
        url = '/security/advanced/blacklistUrls?action=REMOVE_FROM_LIST'
        payload = {
            "blacklistUrls": urls
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    # DLP Policies

    def list_dlpDictionaries(self, dlpDicId=None):
        """
        Gets a list of all DLP Dictionaries.
        :param dlpDicId: type int. dlp dictionary id ( optional parameter)
        """
        if dlpDicId:
            url = f'/dlpDictionaries/{dlpDicId}'
        else:
            url = '/dlpDictionaries'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_dlpDictionaries_lite(self):
        """
        Gets a list of all DLP Dictionary names and ID's only. T
        """

        url = '/dlpDictionaries/lite'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def validateDlpPattern(self, pattern):
        """
        Validates the pattern used by a Pattern and Phrases DLP dictionary type, and provides error information if the
        pattern is invalid.
        :param pattern: Regex pattern
        :return: json
        """
        payload = pattern
        url = '/dlpDictionaries/validateDlpPattern'
        response = self.hp_http.post_call(url, cookies={'JSESSIONID': self.jsessionid}, payload=payload,
                                          error_handling=True)
        return response.json()

    def delete_dlp_dictionaries(self, dlpDicId):
        """
        Deletes the custom DLP category for the specified ID.
        You cannot delete predefined DLP dictionaries.
        You cannot delete a custom dictionary while it is being used by a DLP Engine or policy. Also, predefined
        DLP dictionaries cannot be deleted.
        :param dlpDicId: dlp dictionary ID
        :return: json response
        """
        url = f'/dlpDictionaries/{dlpDicId}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response

    def add_dlpDictionaries(self, dlpdicname, customPhraseMatchType="MATCH_ANY_CUSTOM_PHRASE_PATTERN_DICTIONARY",
                            description=None, phrases=None, patterns=None):
        """
        Adds a new custom DLP dictionary that uses either Patterns and/or Phrases.
        :param dlpdicname: type string.name
        :param phrases: type list. list of phrases
        :phrases valid example:[{
        "action": "PHRASE_COUNT_TYPE_UNIQUE",
        "phrase": "string"
        }, {
        "action": "PHRASE_COUNT_TYPE_UNIQUE",
        "phrase": "string"
        }]
        :param patterns: type list. list of patterns
        :patterns valid example:[{
        "action": "PATTERN_COUNT_TYPE_UNIQUE",
        "phrase": "string"
        }, {
        "action": "PATTERN_COUNT_TYPE_UNIQUE",
        "phrase": "string"
        }]
        :param customPhraseMatchType: type string.customPhraseMatchType
        :param description: description
        """

        if phrases != None:
            for i in phrases:
                if i['action'] not in ["PHRASE_COUNT_TYPE_UNIQUE", "PHRASE_COUNT_TYPE_ALL"]:
                    raise ValueError("Invalid action")
        if patterns != None:
            for k in patterns:
                if k['action'] not in ["PATTERN_COUNT_TYPE_UNIQUE", "PATTERN_COUNT_TYPE_ALL"]:
                    raise ValueError("Invalid action")

        if customPhraseMatchType not in ["MATCH_ALL_CUSTOM_PHRASE_PATTERN_DICTIONARY",
                                         "MATCH_ANY_CUSTOM_PHRASE_PATTERN_DICTIONARY"]:
            raise ValueError("Invalid customPhraseMatchType")

        url = '/dlpDictionaries'
        payload = {
            "name": dlpdicname,
            "description": description,
            "confidenceThreshold": None,
            "customPhraseMatchType": customPhraseMatchType,
            "dictionaryType": "PATTERNS_AND_PHRASES",
            "phrases": phrases,
            "patterns": patterns
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def list_dlpEngines(self, dlpEngineId=None):
        """
        Get a list of DLP engines.
        :param dlpEngineId: type integer. Optinal value. The unique identifier for the DLP engine
        :return: json
        """
        if dlpEngineId:
            url = f'/dlpEngines/{dlpEngineId}'
        else:
            url = '/dlpEngines'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_dlpExactDataMatchSchemas(self):
        """
        Exact Data Match (EDM) templates (or EDM schemas) allow the Zscaler service to identify a record from a structured
        data source that matches predefined criteria. Using the Index Tool, you can create an EDM template that allows
        you to define this criteria (i.e., define the tokens) for your data records by importing a CSV file.
        After the data is defined and submitted within the tool, you can then apply the EDM template to a custom DLP
        dictionary or engine. This endpoint gets the EDM templates for all Index Tools used by the organization

        :return: json
        """
        url = '/dlpExactDataMatchSchemas'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_dlpNotificationTemplates(self, templateId=None):
        """
        Gets a list of DLP notification templates
        :param templateId: type integer. Optional value. The unique identifier  for a DLP notification template
        :return: json
        """
        if templateId:
            url = f'/dlpNotificationTemplates/{templateId}'
        else:

            url = '/dlpNotificationTemplates'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_dlpNotificationTemplates(self, name, subject, plainTextMessage, htmlMessage, attachContent=True,
                                     tlsEnabled=True):
        """

        :param name: type string. The DLP notification template name
        :param subject: type string. The Subject line that is displayed within the DLP notification template
        :param plainTextMessage: type string. The temaplte for the plain text UTF-8 message body that must be displayed
        in the DLP notification email.
        :param htmlMessage: type string. The template for the HTML message body that myst tbe displayed in the DLP
        notification email
        :param attachContent: type boolean. if set to True, the content that is violation is attached to the DLP
        notification email
        :patam tlsEnabled: type boolean. If set to True tls will be used to send email.
        :return:
        """
        url = '/dlpNotificationTemplates'
        payload = {
            "name": name,
            "subject": subject,
            "tlsEnabled": tlsEnabled,
            "attachContent": attachContent,
            "plainTextMessage": plainTextMessage,
            "htmlMessage": htmlMessage
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def delete_dlpNotificationTemplates(self, templateId):
        """
        Deletes a DLP notification template
        :param templateId: type int. the unique identifies for the DLP notification template
        :return: json
        """
        url = f"/dlpNotificationTemplates/{templateId}"
        response = self.hp_http.delete_call(url=url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response

    def list_icapServer(self, icapServerId=None):
        """
        Gets a list of DLP notification templates
        :param icapServerId: type integer. Optional value. The unique identifier for the DLP server using ICAP
        :return: json
        """
        if icapServerId:
            url = f'/icapServers/{icapServerId}'
        else:

            url = '/icapServers'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_idmprofile(self, profileId=None):
        """
        List all the IDM templates for all Index Tools used by the organization. If profileId, it lists the
        IDM template information for the specified ID.
        :param profileId: type integer. Optional value. The unique identifier for the IDM template (or profile)
        :return: json
        """
        if profileId:
            url = f'/idmprofile/{profileId}'
        else:

            url = '/idmprofile'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_webDlpRules(self, ruleId=None):
        """
        list DLP policy rules, excluding SaaS Security API DLP policy rules. If ruleId, list DLP policy rule
        information for the specified ID
        :param ruleId: type integer. Optional value. The unique identifier for theDLP rule
        :return: json
        """
        if ruleId:
            url = f'/webDlpRules/{ruleId}'
        else:

            url = '/webDlpRules'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def delete_webDlpRules(self, ruleId):
        """
        Deletes a DLP policy rule. This endpoint is not applicable to SaaS Security API DLP policy rules.
        :param ruleId: type integer. The unique identifier for the DLP policy rule.
        :return: json
        """
        url = f'/webDlpRules/{ruleId}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=True)
        return response.json()

    # Firewall Policies

    def list_networkServices(self, serviceId=None):
        """
        Gets a list of all network service groups. The search parameters find matching values within the "name" or
        "description" attributes.
        """
        if serviceId:
            url = f'/networkServices/{serviceId}'
        else:
            url = '/networkServices'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_networkServices(self, name, tag=None, srcTcpPorts=None, destTcpPorts=None, srcUdpPorts=None,
                            destUdpPorts=None,
                            type='CUSTOM', description=None, isNameL10nTag=False):

        """
        Adds a new network service.
        :param name: type string. Name
        :param tag: type string
        :param srcTcpPorts: type list. Each element is [{"start": int, "end": int}]
        :param destTcpPorts: type list. Each element is [{"start": int, "end": int}]
        :param srcUdpPorts: type list. Each element is [{"start": int, "end": int}]
        :param destUdpPorts: type list. Each element is [{"start": int, "end": int}]
        :param type: type string. STANDARD|PREDEFINE|CUSTOM
        :param description: type string. Description
        :param isNameL10nTag: type boolean.
        :return: json response from API
        """
        url = '/networkServices'

        payload = {
            "id": 0,
            "name": name,
            "tag": tag,
            "srcTcpPorts": srcTcpPorts,
            "destTcpPorts": destTcpPorts,
            "srcUdpPorts": srcUdpPorts,
            "destUdpPorts": destUdpPorts,
            "type": type,
            "description": description,
            "isNameL10nTag": isNameL10nTag
        }
        print(payload)
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response

    def delete_networkServices(self, serviceid):
        """

        :param serviceid: type int. the unique identifier for the netwokr service
        :return: json
        """
        url = f'/networkServices/{serviceid}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=False)
        return response

    def list_firewallFilteringRules(self, ruleId=None):
        """
        Gets all rules in the Firewall Filtering policy.
        """
        if ruleId:
            url = f'/firewallFilteringRules/{ruleId}'
        else:
            url = '/firewallFilteringRules'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_firewallFilteringRules(self, name, order, state, action, description=None, defaultRule=False,
                                   predefined=False, srcIps=None, destAddresses=None, destIpGroups=None,
                                   srcIpGroups=None, destIpCategories=None, labels=None,nwServices=None, rank=0):
        """
        :param name: type str,  Name of the Firewall Filtering policy rule ["String"]
        :param order: type int, Rule order number of the Firewall Filtering policy rule
        :param state: type str, Possible values : DISABLED or  ENABLED
        :param action: type str, Possible values: ALLOW, BLOCK_DROP, BLOCK_RESET, BLOCK_ICMP, EVAL_NWAPP
        :param rank: type int, Admin rank of the Firewall Filtering policy rule
        :param description: type str, Additional information about the rule
        :param defaultRule: Default is false.If set to true, the default rule is applied
        :param predefined: Boolean
        :param srcIps: type list, List of source IP addresses
        :param destAddresses: type list. List of destination addresses
        :param destIpGroups: type list: List of user-definied destination IP address groups
        :param srcIpGroups: type list: List of user defined source IP address groups
        :param destIpCategories: type list: list of destination IP categories
        :param nwServices: type list. List of user-defined network services on whih the rule is applied
        :return: Default is false.If set to true, a predefined rule is applied
        """

        url = '/firewallFilteringRules'
        payload = {
            "accessControl": "READ_WRITE",
            "enableFullLogging": False,
            "name": name,
            "order": order,
            "rank": rank,
            "action": action,
            "state": state,
            "predefined": predefined,
            "defaultRule": defaultRule,
            "description": description,

        }
        if srcIps:
            payload.update(srcIps=srcIps)
        if srcIpGroups:
            payload.update(srcIpGroups=srcIpGroups)
        if destAddresses:
            payload.update(destAddresses=destAddresses)
        if destIpGroups:
            payload.update(destIpGroups=destIpGroups)
        if labels:
            payload.update(labels=labels)
        if destIpCategories:
            payload.update(destIpCategories=destIpCategories)
        if nwServices:
            payload.update(nwServices=nwServices)
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response

    def delete_firewallFIlteringRules(self, ruleId):

        """
        Deletes a Firewall Filtering policy rule for the specified ID.

        :param ruleId: type integer: The unique identifier for the policy rule
        :return: json
        """
        url = f'/firewallFilteringRules/{ruleId}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid},
                                            error_handling=False)
        return response

    def list_ipSourceGroups(self, ipGroupId=None):
        """
        Gets a list of all IP source groups. The search parameters find matching values within the "name" or
        "description" attributes.
        :param ipGroupId: Option ip group id
        """
        if ipGroupId:
            url = f'/ipSourceGroups/{ipGroupId}'
        else:
            url = '/ipSourceGroups'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_ipSourceGroups_lite(self, ):
        """
        Gets a name and ID dictionary of all IP source groups
        :return:
        """
        url = '/ipSourceGroups/lite'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_ipDestinationGroups(self, ipGroupId=None):
        """
        Gets a list of all IP source groups. The search parameters find matching values within the "name" or
        "description" attributes.
        :param ipGroupId: Option ip group id
        """
        if ipGroupId:
            url = f'/ipDestinationGroups/{ipGroupId}'
        else:
            url = '/ipDestinationGroups/'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_ipDestinationGroups_lite(self):
        """
        Gets a name and ID dictionary of all IP destination groups
        return json
        """

        url = '/ipDestinationGroups/lite'
        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_ipSourceGroups(self, name, ipAddresses, description=None):
        """
        :param name: mame
        :param ipAddresses: list of IP addresses
        :param description: description
        """

        url = '/ipSourceGroups'
        payload = {
            "name": name,
            "ipAddresses": ipAddresses,
            "description": description
        }
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    def delete_ipSourceGroups(self, ipGroupId):
        """
        Deletes the IP source group for the specified ID
        :param ipGroupId:  type int. The uniquye identifies for the IP source group
        :return: json
        """
        url = f'/ipSourceGroups/{ipGroupId}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True,
                                            payload={})
        return response

    def delete_ipDestinationGroups(self, ipGroupId):
        """
        Deletes the IP destination group for the specified ID
        :param ipGroupId:  type int. The uniquye identifies for the IP source group
        :return: json
        """
        url = f'/ipDestinationGroups/{ipGroupId}'
        response = self.hp_http.delete_call(url, cookies={'JSESSIONID': self.jsessionid}, error_handling=True,
                                            payload={})
        return response

    def add_ipDestinationGroups(self, name, type, addresses, ipCategories=None, countries=None, description=None):
        """
        :param name: mame
        :param type: Destination IP group type. Either DSTN_IP or  DSTN_FQDN or DSTN_DOMAIN
        :param addresses: List of Destination IP addresses within the group.
        :param description: description
        :param ipCategories: List of Destination IP address URL categories. You can identify destination based
        on the URL category of the domain. Default value ANY
        :param countries: list of destination IP address countries. You can identify destinations based on the location
        of a server.Default value ANY
        """
        if type not in ["DSTN_IP", "DSTN_FQDN", "DSTN_DOMAIN"]:
            raise ValueError("Invalid destination type ")
        if countries:
            for i in countries:
                if i not in valid_countries:
                    raise ValueError("Invalid country ")
        else:
            countries = []

        if ipCategories:
            for j in ipCategories:
                if j not in valid_category_ids:
                    raise ValueError("Invalid country ")
        else:
            ipCategories = []

        url = '/ipDestinationGroups'
        payload = {
            "name": name,
            "type": type,
            "addresses": addresses,
            "ipCategories": ipCategories,
            "countries": countries,
            "description": description
        }
        print(payload)
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()

    # Device Groups

    def list_devices_groups(self, query=None):
        """
        Gets a list of device groups
        :param query:
        :return: List
        """
        if query:
            url = f"/deviceGroups?{query}"
        else:
            url = "/deviceGroups"

        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def list_devices(self, query=None):
        """
        Gets a list of devices. Any given search parameters will be applied during device search. Search parameters are
         based on device name, model, owner, OS type, and OS version. The devices listed can also be restricted to return
         information only for ones belonging to specific users.
        :param query:
        :return: List
        """
        if query:
            url = f"/deviceGroups/devices?{query}"
        else:
            url = "/deviceGroups/devices"

        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    # Rule Labels
    def list_rule_labels(self, ruleLabelId=None):
        """
        Gets rule label information for the specified ID
        :param ruleLabelId:
        :return: List
        """
        if ruleLabelId:
            url = f"/ruleLabels/{ruleLabelId}"

        else:
            url = "/ruleLabels?pageSize=1000"

        response = self.hp_http.get_call(url, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def update_call(self, url, payload):
        """
        Generic PUT call. This call will overwrite all the configuration with the new payload
        :param url: url of Zscaler API call
        :param payload: type json. Payload
        """
        response = self.hp_http.put_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                         error_handling=True)
        return response.json()

    def add_call(self, url, payload):
        """
        Generic POST call. This call will add all the configuration with the new payload
        :param url: url of Zscaler API call
        :param payload: type json. Payload
        """
        response = self.hp_http.post_call(url, payload=payload, cookies={'JSESSIONID': self.jsessionid},
                                          error_handling=True)
        return response.json()



import pdb

import requests

from helpers.http_calls import HttpCalls


class ZpaPortalTalker(object):

    def __init__(self, customerId, cloud='https://api.private.zscaler.com'):
        self.base_uri = cloud
        self.version = '1.0'
        self.cookies = None
        self.bear = None
        self.token = None
        self.headers = None
        self.customerId = customerId
        self.hp_http = HttpCalls(host=self.base_uri, verify=True)

    def _obtain_all_pages(self, url):
        result = []
        response = requests.request("GET", url, headers=self.token)
        if int(response.json()['totalPages']) > 1:
            i = 1
            while i <= int(response.json()['totalPages']):
                result = result + requests.request("GET", url, headers=self.token).json()['list']
                i += 1
        else:
            result = response.json()['list']
        return result

    def authenticate(self, username, password):
        """
        Method to obtain authorization token for subsequent calls.
        :param username: Email address
        :param password: Password for given user
        """
        url = "/base/api/zpa/signin"
        payload = {'username': username,
                   'password': password}
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = self.hp_http.post_call(url=url, payload=payload, headers=headers, urlencoded=True)
        self.token = response.json()['Z-AUTH-TOKEN']
        self.headers = {
            'Content-Type': 'application/json',
            'Accept': '*/*',
            'Accept-Encoding': 'gzip, deflate, br',
            'Authorization': 'Bearer ' + self.token
        }
        return


    def list_admin_users(self):
        """List admins users"""
        url = f'/shift/api/v2/admin/customers/{self.customerId}/users'
        response = self.hp_http.get_call(url=url, headers=self.headers)

        if int(response.json()['totalPages']) > 1:
            response = self._obtain_all_pages(url)
        else:
            response = response.json()['list']
        return response

    def list_admin_roles(self):
        """List admins roles"""
        url = f'/zpn/api/v1/admin/customers/{self.customerId}/roles'
        response = self.hp_http.get_call(url=url, headers=self.headers)
        return response



from setuptools import setup, find_packages

setup(
    name='zscaler_api_talkers',
    version='3.10',
    author='Sergio Pereira',
    author_email='sergitopereira@hotmail.com',
    packages=find_packages(),
    url='https://github.com/sergitopereira/zscaler_api_talkers.git',
    license='LICENSE.txt',
    description='Unofficial Zscaler API python SDK for ZIA, ZPA and ZCC',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    install_requires=[
        "requests",
        "appdirs",
        "ipython"
    ],
)



valid_category_ids = ["ANY", "NONE", "OTHER_ADULT_MATERIAL", "ADULT_THEMES", "LINGERIE_BIKINI", "NUDITY",
                      "PORNOGRAPHY",
                      "SEXUALITY", "ADULT_SEX_EDUCATION", "K_12_SEX_EDUCATION", "SOCIAL_ADULT",
                      "OTHER_BUSINESS_AND_ECONOMY",
                      "CORPORATE_MARKETING", "FINANCE", "PROFESSIONAL_SERVICES", "CLASSIFIEDS",
                      "TRADING_BROKARAGE_INSURANCE",
                      "CUSTOM_00", "CUSTOM_01", "CUSTOM_02", "CUSTOM_03", "CUSTOM_04", "CUSTOM_05", "CUSTOM_06",
                      "CUSTOM_07",
                      "CUSTOM_08", "CUSTOM_09", "CUSTOM_10", "CUSTOM_11", "CUSTOM_12", "CUSTOM_13", "CUSTOM_14",
                      "CUSTOM_15",
                      "CUSTOM_16", "CUSTOM_17", "CUSTOM_18", "CUSTOM_19", "CUSTOM_20", "CUSTOM_21", "CUSTOM_22",
                      "CUSTOM_23",
                      "CUSTOM_24", "CUSTOM_25", "CUSTOM_26", "CUSTOM_27", "CUSTOM_28", "CUSTOM_29", "CUSTOM_30",
                      "CUSTOM_31",
                      "CUSTOM_32", "CUSTOM_33", "CUSTOM_34", "CUSTOM_35", "CUSTOM_36", "CUSTOM_37", "CUSTOM_38",
                      "CUSTOM_39",
                      "CUSTOM_40", "CUSTOM_41", "CUSTOM_42", "CUSTOM_43", "CUSTOM_44", "CUSTOM_45", "CUSTOM_46",
                      "CUSTOM_47",
                      "CUSTOM_48", "CUSTOM_49", "CUSTOM_50", "CUSTOM_51", "CUSTOM_52", "CUSTOM_53", "CUSTOM_54",
                      "CUSTOM_55",
                      "CUSTOM_56", "CUSTOM_57", "CUSTOM_58", "CUSTOM_59", "CUSTOM_60", "CUSTOM_61", "CUSTOM_62",
                      "CUSTOM_63",
                      "CUSTOM_64", "CUSTOM_65", "CUSTOM_66", "CUSTOM_67", "CUSTOM_68", "CUSTOM_69", "CUSTOM_70",
                      "CUSTOM_71",
                      "CUSTOM_72", "CUSTOM_73", "CUSTOM_74", "CUSTOM_75", "CUSTOM_76", "CUSTOM_77", "CUSTOM_78",
                      "CUSTOM_79",
                      "CUSTOM_80", "CUSTOM_81", "CUSTOM_82", "CUSTOM_83", "CUSTOM_84", "CUSTOM_85", "CUSTOM_86",
                      "CUSTOM_87",
                      "CUSTOM_88", "CUSTOM_89", "CUSTOM_90", "CUSTOM_91", "CUSTOM_92", "CUSTOM_93", "CUSTOM_94",
                      "CUSTOM_95",
                      "CUSTOM_96", "CUSTOM_97", "CUSTOM_98", "CUSTOM_99", "CUSTOM_100", "CUSTOM_101",
                      "CUSTOM_102", "CUSTOM_103",
                      "CUSTOM_104", "CUSTOM_105", "CUSTOM_106", "CUSTOM_107", "CUSTOM_108", "CUSTOM_109",
                      "CUSTOM_110",
                      "CUSTOM_111", "CUSTOM_112", "CUSTOM_113", "CUSTOM_114", "CUSTOM_115", "CUSTOM_116",
                      "CUSTOM_117",
                      "CUSTOM_118", "CUSTOM_119", "CUSTOM_120", "CUSTOM_121", "CUSTOM_122", "CUSTOM_123",
                      "CUSTOM_124",
                      "CUSTOM_125", "CUSTOM_126", "CUSTOM_127", "CUSTOM_128", "CUSTOM_129", "CUSTOM_130",
                      "CUSTOM_131",
                      "CUSTOM_132", "CUSTOM_133", "CUSTOM_134", "CUSTOM_135", "CUSTOM_136", "CUSTOM_137",
                      "CUSTOM_138",
                      "CUSTOM_139", "CUSTOM_140", "CUSTOM_141", "CUSTOM_142", "CUSTOM_143", "CUSTOM_144",
                      "CUSTOM_145",
                      "CUSTOM_146", "CUSTOM_147", "CUSTOM_148", "CUSTOM_149", "CUSTOM_150", "CUSTOM_151",
                      "CUSTOM_152",
                      "CUSTOM_153", "CUSTOM_154", "CUSTOM_155", "CUSTOM_156", "CUSTOM_157", "CUSTOM_158",
                      "CUSTOM_159",
                      "CUSTOM_160", "CUSTOM_161", "CUSTOM_162", "CUSTOM_163", "CUSTOM_164", "CUSTOM_165",
                      "CUSTOM_166",
                      "CUSTOM_167", "CUSTOM_168", "CUSTOM_169", "CUSTOM_170", "CUSTOM_171", "CUSTOM_172",
                      "CUSTOM_173",
                      "CUSTOM_174", "CUSTOM_175", "CUSTOM_176", "CUSTOM_177", "CUSTOM_178", "CUSTOM_179",
                      "CUSTOM_180",
                      "CUSTOM_181", "CUSTOM_182", "CUSTOM_183", "CUSTOM_184", "CUSTOM_185", "CUSTOM_186",
                      "CUSTOM_187",
                      "CUSTOM_188", "CUSTOM_189", "CUSTOM_190", "CUSTOM_191", "CUSTOM_192", "CUSTOM_193",
                      "CUSTOM_194",
                      "CUSTOM_195", "CUSTOM_196", "CUSTOM_197", "CUSTOM_198", "CUSTOM_199", "CUSTOM_200",
                      "CUSTOM_201",
                      "CUSTOM_202", "CUSTOM_203", "CUSTOM_204", "CUSTOM_205", "CUSTOM_206", "CUSTOM_207",
                      "CUSTOM_208",
                      "CUSTOM_209", "CUSTOM_210", "CUSTOM_211", "CUSTOM_212", "CUSTOM_213", "CUSTOM_214",
                      "CUSTOM_215",
                      "CUSTOM_216", "CUSTOM_217", "CUSTOM_218", "CUSTOM_219", "CUSTOM_220", "CUSTOM_221",
                      "CUSTOM_222",
                      "CUSTOM_223", "CUSTOM_224", "CUSTOM_225", "CUSTOM_226", "CUSTOM_227", "CUSTOM_228",
                      "CUSTOM_229",
                      "CUSTOM_230", "CUSTOM_231", "CUSTOM_232", "CUSTOM_233", "CUSTOM_234", "CUSTOM_235",
                      "CUSTOM_236",
                      "CUSTOM_237", "CUSTOM_238", "CUSTOM_239", "CUSTOM_240", "CUSTOM_241", "CUSTOM_242",
                      "CUSTOM_243",
                      "CUSTOM_244", "CUSTOM_245", "CUSTOM_246", "CUSTOM_247", "CUSTOM_248", "CUSTOM_249",
                      "CUSTOM_250",
                      "CUSTOM_251", "CUSTOM_252", "CUSTOM_253", "CUSTOM_254", "CUSTOM_255", "CUSTOM_256",
                      "OTHER_DRUGS",
                      "MARIJUANA", "OTHER_EDUCATION", "CONTINUING_EDUCATION_COLLEGES", "HISTORY", "K_12",
                      "REFERENCE_SITES",
                      "SCIENCE_AND_TECHNOLOGY", "OTHER_ENTERTAINMENT_AND_RECREATION", "ENTERTAINMENT",
                      "TELEVISION_AND_MOVIES",
                      "MUSIC", "STREAMING_MEDIA", "RADIO_STATIONS", "GAMBLING", "OTHER_GAMES",
                      "SOCIAL_NETWORKING_GAMES",
                      "OTHER_GOVERNMENT_AND_POLITICS", "GOVERNMENT", "POLITICS", "HEALTH",
                      "OTHER_ILLEGAL_OR_QUESTIONABLE",
                      "COPYRIGHT_INFRINGEMENT", "COMPUTER_HACKING", "QUESTIONABLE", "PROFANITY", "MATURE_HUMOR",
                      "ANONYMIZER",
                      "OTHER_INFORMATION_TECHNOLOGY", "TRANSLATORS", "IMAGE_HOST", "FILE_HOST",
                      "SHAREWARE_DOWNLOAD",
                      "WEB_BANNERS", "WEB_HOST", "WEB_SEARCH", "PORTALS", "SAFE_SEARCH_ENGINE", "CDN",
                      "OSS_UPDATES",
                      "DNS_OVER_HTTPS", "OTHER_INTERNET_COMMUNICATION", "INTERNET_SERVICES",
                      "DISCUSSION_FORUMS",
                      "ONLINE_CHAT",
                      "EMAIL_HOST", "BLOG", "P2P_COMMUNICATION", "REMOTE_ACCESS", "WEB_CONFERENCING",
                      "ZSPROXY_IPS",
                      "JOB_SEARCH", "MILITANCY_HATE_AND_EXTREMISM", "OTHER_MISCELLANEOUS",
                      "MISCELLANEOUS_OR_UNKNOWN",
                      "NEWLY_REG_DOMAINS", "NON_CATEGORIZABLE", "NEWS_AND_MEDIA", "OTHER_RELIGION",
                      "TRADITIONAL_RELIGION",
                      "CULT", "ALT_NEW_AGE", "OTHER_SECURITY", "ADWARE_OR_SPYWARE", "ENCR_WEB_CONTENT",
                      "MALICIOUS_TLD",
                      "OTHER_SHOPPING_AND_AUCTIONS", "SPECIALIZED_SHOPPING", "REAL_ESTATE", "ONLINE_AUCTIONS",
                      "OTHER_SOCIAL_AND_FAMILY_ISSUES", "SOCIAL_ISSUES", "FAMILY_ISSUES",
                      "OTHER_SOCIETY_AND_LIFESTYLE",
                      "ART_CULTURE", "ALTERNATE_LIFESTYLE", "HOBBIES_AND_LEISURE", "DINING_AND_RESTAURANT",
                      "ALCOHOL_TOBACCO",
                      "SOCIAL_NETWORKING", "SPECIAL_INTERESTS", "SPORTS", "TASTELESS", "TRAVEL", "USER_DEFINED",
                      "VEHICLES",
                      "VIOLENCE", "WEAPONS_AND_BOMBS"]

super_categories = ["ENTERTAINMENT", "MUSIC", "OTHER_ENTERTAINMENT_AND_RECREATION", "RADIO_STATIONS",
                    "STREAMING_MEDIA", "TELEVISION_AND_MOVIES", "NEWS_AND_MEDIA", "USER_DEFINED", "CLASSIFIEDS",
                    "CORPORATE_MARKETING", "FINANCE", "TRADING_BROKERAGE_INSURANCE", "OTHER_BUSINESS_AND_ECONOMY",
                    "PROFESSIONAL_SERVICES", "CONTINUING_EDUCATION_COLLEGES", "HISTORY", "K_12", "OTHER_EDUCATION",
                    "REFERENCE_SITES", "SCIENCE_AND_TECHNOLOGY", "WEB_BANNERS", "CDN", "DNS_OVER_HTTPS", "FILE_HOST",
                    "IMAGE_HOST", "OSS_UPDATES", "OTHER_INFORMATION_TECHNOLOGY", "PORTALS", "SAFE_SEARCH_ENGINE",
                    "SHAREWARE_DOWNLOAD", "TRANSLATORS", "WEB_HOST", "WEB_SEARCH", "BLOG", "DISCUSSION_FORUMS",
                    "INTERNET_SERVICES", "ONLINE_CHAT", "OTHER_INTERNET_COMMUNICATION", "P2P_COMMUNICATION",
                    "REMOTE_ACCESS", "EMAIL_HOST", "WEB_CONFERENCING", "ZSPROXY_IPS", "JOB_SEARCH", "GOVERNMENT",
                    "OTHER_GOVERNMENT_AND_POLITICS", "POLITICS", "MISCELLANEOUS_OR_UNKNOWN", "NEWLY_REG_DOMAINS",
                    "NON_CATEGORIZABLE", "OTHER_MISCELLANEOUS", "TRAVEL", "VEHICLES", "ADULT_SEX_EDUCATION",
                    "ADULT_THEMES", "K_12_SEX_EDUCATION", "LINGERIE_BIKINI", "NUDITY", "OTHER_ADULT_MATERIAL",
                    "PORNOGRAPHY", "SEXUALITY", "SOCIAL_ADULT", "DRUGS", "MARIJUANA", "GAMBLING", "ANONYMIZER",
                    "COMPUTER_HACKING", "COPYRIGHT_INFRINGEMENT", "MATURE_HUMOR", "OTHER_ILLEGAL_OR_QUESTIONABLE",
                    "PROFANITY", "QUESTIONABLE", "MILITANCY_HATE_AND_EXTREMISM", "TASTELESS", "VIOLENCE",
                    "WEAPONS_AND_BOMBS", "OTHER_GAMES", "SOCIAL_NETWORKING_GAMES", "HEALTH", "ALT_NEW_AGE", "CULT",
                    "OTHER_RELIGION", "TRADITIONAL_RELIGION", "ONLINE_AUCTIONS", "SPECIALIZED_SHOPPING",
                    "OTHER_SHOPPING_AND_AUCTIONS", "REAL_ESTATE", "FAMILY_ISSUES", "OTHER_SOCIAL_AND_FAMILY_ISSUES",
                    "SOCIAL_ISSUES", "ALCOHOL_TOBACCO", "LIFESTYLE", "ART_CULTURE", "DINING_AND_RESTAURANT",
                    "HOBBIES_AND_LEISURE", "OTHER_SOCIETY_AND_LIFESTYLE", "SOCIAL_NETWORKING", "SPECIAL_INTERESTS",
                    "SPORTS", "ENCR_WEB_CONTENT", "MALICIOUS_TLD", "OTHER_SECURITY", "ADWARE_OR_SPYWARE"]

valid_countries = ["ANY", "NONE", "COUNTRY_AD", "COUNTRY_AE", "COUNTRY_AF", "COUNTRY_AG", "COUNTRY_AI", "COUNTRY_AL",
                   "COUNTRY_AM", "COUNTRY_AN", "COUNTRY_AO", "COUNTRY_AQ", "COUNTRY_AR", "COUNTRY_AS", "COUNTRY_AT",
                   "COUNTRY_AU", "COUNTRY_AW", "COUNTRY_AZ", "COUNTRY_BA", "COUNTRY_BB", "COUNTRY_BD", "COUNTRY_BE",
                   "COUNTRY_BF", "COUNTRY_BG", "COUNTRY_BH", "COUNTRY_BI", "COUNTRY_BJ", "COUNTRY_BM", "COUNTRY_BN",
                   "COUNTRY_BO", "COUNTRY_BR", "COUNTRY_BS", "COUNTRY_BT", "COUNTRY_BV", "COUNTRY_BW", "COUNTRY_BY",
                   "COUNTRY_BZ", "COUNTRY_CA", "COUNTRY_CC", "COUNTRY_CD", "COUNTRY_CF", "COUNTRY_CG", "COUNTRY_CH",
                   "COUNTRY_CI", "COUNTRY_CK", "COUNTRY_CL", "COUNTRY_CM", "COUNTRY_CN", "COUNTRY_CO", "COUNTRY_CR",
                   "COUNTRY_CU", "COUNTRY_CV", "COUNTRY_CX", "COUNTRY_CY", "COUNTRY_CZ", "COUNTRY_DE", "COUNTRY_DJ",
                   "COUNTRY_DK", "COUNTRY_DM", "COUNTRY_DO", "COUNTRY_DZ", "COUNTRY_EC", "COUNTRY_EE", "COUNTRY_EG",
                   "COUNTRY_EH", "COUNTRY_ER", "COUNTRY_ES", "COUNTRY_ET", "COUNTRY_FI", "COUNTRY_FJ", "COUNTRY_FK",
                   "COUNTRY_FM", "COUNTRY_FO", "COUNTRY_FR", "COUNTRY_FX", "COUNTRY_GA", "COUNTRY_GB", "COUNTRY_GD",
                   "COUNTRY_GE", "COUNTRY_GF", "COUNTRY_GH", "COUNTRY_GI", "COUNTRY_GL", "COUNTRY_GM", "COUNTRY_GN",
                   "COUNTRY_GP", "COUNTRY_GQ", "COUNTRY_GR", "COUNTRY_GS", "COUNTRY_GT", "COUNTRY_GU", "COUNTRY_GW",
                   "COUNTRY_GY", "COUNTRY_HK", "COUNTRY_HM", "COUNTRY_HN", "COUNTRY_HR", "COUNTRY_HT", "COUNTRY_HU",
                   "COUNTRY_ID", "COUNTRY_IE", "COUNTRY_IL", "COUNTRY_IN", "COUNTRY_IO", "COUNTRY_IQ", "COUNTRY_IR",
                   "COUNTRY_IS", "COUNTRY_IT", "COUNTRY_JM", "COUNTRY_JO", "COUNTRY_JP", "COUNTRY_KE", "COUNTRY_KG",
                   "COUNTRY_KH", "COUNTRY_KI", "COUNTRY_KM", "COUNTRY_KN", "COUNTRY_KP", "COUNTRY_KR", "COUNTRY_KW",
                   "COUNTRY_KY", "COUNTRY_KZ", "COUNTRY_LA", "COUNTRY_LB", "COUNTRY_LC", "COUNTRY_LI", "COUNTRY_LK",
                   "COUNTRY_LR", "COUNTRY_LS", "COUNTRY_LT", "COUNTRY_LU", "COUNTRY_LV", "COUNTRY_LY", "COUNTRY_MA",
                   "COUNTRY_MC", "COUNTRY_MD", "COUNTRY_MG", "COUNTRY_MH", "COUNTRY_MK", "COUNTRY_ML", "COUNTRY_MM",
                   "COUNTRY_MN", "COUNTRY_MO", "COUNTRY_MP", "COUNTRY_MQ", "COUNTRY_MR", "COUNTRY_MS", "COUNTRY_MT",
                   "COUNTRY_MU", "COUNTRY_MV", "COUNTRY_MW", "COUNTRY_MX", "COUNTRY_MY", "COUNTRY_MZ", "COUNTRY_NA",
                   "COUNTRY_NC", "COUNTRY_NE", "COUNTRY_NF", "COUNTRY_NG", "COUNTRY_NI", "COUNTRY_NL", "COUNTRY_NO",
                   "COUNTRY_NP", "COUNTRY_NR", "COUNTRY_NU", "COUNTRY_NZ", "COUNTRY_OM", "COUNTRY_PA", "COUNTRY_PE",
                   "COUNTRY_PF", "COUNTRY_PG", "COUNTRY_PH", "COUNTRY_PK", "COUNTRY_PL", "COUNTRY_PM", "COUNTRY_PN",
                   "COUNTRY_PR", "COUNTRY_PS", "COUNTRY_PT", "COUNTRY_PW", "COUNTRY_PY", "COUNTRY_QA", "COUNTRY_RE",
                   "COUNTRY_RO", "COUNTRY_RU", "COUNTRY_RW", "COUNTRY_SA", "COUNTRY_SB", "COUNTRY_SC", "COUNTRY_SD",
                   "COUNTRY_SE", "COUNTRY_SG", "COUNTRY_SH", "COUNTRY_SI", "COUNTRY_SJ", "COUNTRY_SK", "COUNTRY_SL",
                   "COUNTRY_SM", "COUNTRY_SN", "COUNTRY_SO", "COUNTRY_SR", "COUNTRY_ST", "COUNTRY_SV", "COUNTRY_SY",
                   "COUNTRY_SZ", "COUNTRY_TC", "COUNTRY_TD", "COUNTRY_TF", "COUNTRY_TG", "COUNTRY_TH", "COUNTRY_TJ",
                   "COUNTRY_TK", "COUNTRY_TM", "COUNTRY_TN", "COUNTRY_TO", "COUNTRY_TL", "COUNTRY_TR", "COUNTRY_TT",
                   "COUNTRY_TV", "COUNTRY_TW", "COUNTRY_TZ", "COUNTRY_UA", "COUNTRY_UG", "COUNTRY_UM", "COUNTRY_US",
                   "COUNTRY_UY", "COUNTRY_UZ", "COUNTRY_VA", "COUNTRY_VC", "COUNTRY_VE", "COUNTRY_VG", "COUNTRY_VI",
                   "COUNTRY_VN", "COUNTRY_VU", "COUNTRY_WF", "COUNTRY_WS", "COUNTRY_YE", "COUNTRY_YT", "COUNTRY_RS",
                   "COUNTRY_ZA", "COUNTRY_ZM", "COUNTRY_ME", "COUNTRY_ZW", "COUNTRY_AX", "COUNTRY_GG", "COUNTRY_IM",
                   "COUNTRY_JE", "COUNTRY_BL", "COUNTRY_MF"]
